use std::rc::Rc;

use super::{
    environment::Environment,
    evaluate::{EvaluateExpr, EvaluateValue},
};
use crate::parser::syntax::{BinaryOp, Expr, UnaryOp, Value};

pub trait ValueCompute {
    fn add(&self, right: &Value) -> Result<Value, String>;
    fn sub(&self, right: &Value) -> Result<Value, String>;
    fn mult(&self, right: &Value) -> Result<Value, String>;
    fn div(&self, right: &Value) -> Result<Value, String>;
    fn modulo(&self, right: &Value) -> Result<Value, String>;
    fn equal(&self, right: &Value) -> Result<Value, String>;
    fn not_equal(&self, right: &Value) -> Result<Value, String>;
    fn greater_equal(&self, right: &Value) -> Result<Value, String>;
    fn greater_than(&self, right: &Value) -> Result<Value, String>;
    fn lesser_equal(&self, right: &Value) -> Result<Value, String>;
    fn lesser_than(&self, right: &Value) -> Result<Value, String>;
}

impl ValueCompute for Value {
    fn add(&self, right: &Value) -> Result<Value, String> {
        match (self, right) {
            (Value::Number(l), Value::Number(r)) => match (*l).checked_add(*r) {
                Some(n) => Ok(Value::Number(n)),
                None => Err(format!("Integer overflow: {} + {}", l, r)),
            },
            (Value::String(l), Value::Char(r)) => {
                let mut copy = Vec::with_capacity(l.len() + 1);
                copy.extend_from_slice(l);
                copy.push(*r);
                Ok(Value::String(Rc::new(copy)))
            }
            (Value::Char(l), Value::String(r)) => {
                let mut copy = Vec::with_capacity(r.len() + 1);
                copy.push(*l);
                copy.extend_from_slice(r);
                Ok(Value::String(Rc::new(copy)))
            }
            (Value::Char(l), Value::Char(r)) => Ok(Value::Char(l + r)),
            (Value::Char(l), Value::Number(r)) => match l.checked_add(*r as u8) {
                Some(n) => Ok(Value::Char(n)),
                None => Err(format!("Char overflow: {} + {}", char::from(*l), r)),
            },
            (Value::Number(l), Value::Char(r)) => match l.checked_add(*r as i64) {
                Some(n) => Ok(Value::Number(n)),
                None => Err(format!("Integer overflow: {} + {}", l, char::from(*r))),
            },
            (Value::String(l), Value::String(r)) => {
                let new = l.iter().chain(r.iter()).copied().collect();
                Ok(Value::String(Rc::new(new)))
            }
            (Value::Array(l), Value::Array(r)) => {
                let new = l.iter().chain(r.iter()).cloned().collect();
                Ok(Value::Array(Rc::new(new)))
            }
            (Value::Array(l), r) => {
                let mut new = (**l).clone();
                new.push(r.clone());
                Ok(Value::Array(Rc::new(new)))
            }
            (_, _) => Err(format!(
                "Cannot add {} and {}",
                self.get_type(),
                right.get_type()
            )),
        }
    }

    fn sub(&self, right: &Value) -> Result<Value, String> {
        match (self, right) {
            (Value::Number(l), Value::Number(r)) => match l.checked_sub(*r) {
                Some(n) => Ok(Value::Number(n)),
                None => Err(format!("Integer overflow: {} - {}", l, r)),
            },
            (Value::Char(l), Value::Char(r)) => match l.checked_sub(*r) {
                Some(n) => Ok(Value::Char(n)),
                None => Err(format!(
                    "Char overflow: {} - {}",
                    char::from(*l),
                    char::from(*r)
                )),
            },
            (Value::Char(l), Value::Number(r)) => match l.checked_sub(*r as u8) {
                Some(n) => Ok(Value::Char(n)),
                None => Err(format!("Char overflow: {} - {}", char::from(*l), r)),
            },
            (Value::Number(l), Value::Char(r)) => match l.checked_sub(*r as i64) {
                Some(n) => Ok(Value::Number(n)),
                None => Err(format!("Integer overflow: {} - {}", l, char::from(*r))),
            },
            (_, _) => Err(format!(
                "Cannot subtract {} and {}",
                self.get_type(),
                right.get_type()
            )),
        }
    }

    fn mult(&self, right: &Value) -> Result<Value, String> {
        match (self, right) {
            (Value::Number(l), Value::Number(r)) => match l.checked_mul(*r) {
                Some(n) => Ok(Value::Number(n)),
                None => Err(format!("Integer overflow: {} * {}", l, r)),
            },
            (Value::Char(c), Value::Number(n)) | (Value::Number(n), Value::Char(c)) => Ok(
                Value::String(Rc::new((0..*n).map(|_| *c).collect::<Vec<u8>>())),
            ),
            (Value::String(s), Value::Number(n)) | (Value::Number(n), Value::String(s)) => {
                Ok(Value::String(Rc::new(
                    (0..*n).flat_map(|_| s.iter().copied()).collect::<Vec<u8>>(),
                )))
            }
            (Value::Array(arr), Value::Number(n)) | (Value::Number(n), Value::Array(arr)) => {
                Ok(Value::Array(Rc::new(
                    (0..*n)
                        .flat_map(|_| arr.iter().cloned())
                        .collect::<Vec<Value>>(),
                )))
            }
            (_, _) => Err(format!(
                "Cannot multiply {} and {}",
                self.get_type(),
                right.get_type()
            )),
        }
    }

    fn div(&self, right: &Value) -> Result<Value, String> {
        match (self, right) {
            (Value::Number(l), Value::Number(r)) => {
                if *r == 0 {
                    Err("attempt to divide by zero".to_string())
                } else {
                    Ok(Value::Number(l / r))
                }
            }
            (_, _) => Err(format!(
                "Cannot divide {} and {}",
                self.get_type(),
                right.get_type()
            )),
        }
    }

    fn modulo(&self, right: &Value) -> Result<Value, String> {
        match (self, right) {
            (Value::Number(l), Value::Number(r)) => {
                if *r == 0 {
                    Err("attempt to calculate the remainder with a divisor of zero".to_string())
                } else {
                    Ok(Value::Number(l % r))
                }
            }
            (_, _) => Err(format!(
                "Cannot modulo {} and {}",
                self.get_type(),
                right.get_type()
            )),
        }
    }

    fn equal(&self, right: &Value) -> Result<Value, String> {
        match (self, right) {
            (Value::Number(l), Value::Number(r)) => Ok(Value::Boolean(l == r)),
            (Value::Boolean(l), Value::Boolean(r)) => Ok(Value::Boolean(l == r)),
            (Value::String(l), Value::String(r)) => Ok(Value::Boolean(l == r)),
            (Value::Char(l), Value::Char(r)) => Ok(Value::Boolean(l == r)),
            (Value::Array(l), Value::Array(r)) => {
                if l.len() != r.len() {
                    Ok(Value::Boolean(false))
                } else {
                    for pair in l.iter().zip(r.iter()) {
                        if let Value::Boolean(false) = pair.0.equal(pair.1)? {
                            return Ok(Value::Boolean(false));
                        }
                    }
                    Ok(Value::Boolean(true))
                }
            }
            (l, r) => Err(format!("Cannot evaluate equality between {} and {}", l, r)),
        }
    }

    fn not_equal(&self, right: &Value) -> Result<Value, String> {
        match (self, right) {
            (Value::Number(l), Value::Number(r)) => Ok(Value::Boolean(l != r)),
            (Value::Boolean(l), Value::Boolean(r)) => Ok(Value::Boolean(l != r)),
            (Value::String(l), Value::String(r)) => Ok(Value::Boolean(l != r)),
            (Value::Char(l), Value::Char(r)) => Ok(Value::Boolean(l != r)),
            (Value::Array(l), Value::Array(r)) => {
                if l.len() != r.len() {
                    Ok(Value::Boolean(true))
                } else {
                    for pair in l.iter().zip(r.iter()) {
                        if let Value::Boolean(true) = pair.0.equal(pair.1)? {
                            return Ok(Value::Boolean(true));
                        }
                    }
                    Ok(Value::Boolean(false))
                }
            }
            (l, r) => Err(format!(
                "Cannot evaluate inequality between {} and {}",
                l, r
            )),
        }
    }

    fn greater_equal(&self, right: &Value) -> Result<Value, String> {
        match (self, right) {
            (Value::Number(l), Value::Number(r)) => Ok(Value::Boolean(l >= r)),
            (Value::String(l), Value::String(r)) => Ok(Value::Boolean(l >= r)),
            (Value::Char(l), Value::Char(r)) => Ok(Value::Boolean(l >= r)),
            (l, r) => Err(format!("Cannot compare {} and {}", l, r)),
        }
    }

    fn greater_than(&self, right: &Value) -> Result<Value, String> {
        match (self, right) {
            (Value::Number(l), Value::Number(r)) => Ok(Value::Boolean(l > r)),
            (Value::String(l), Value::String(r)) => Ok(Value::Boolean(l > r)),
            (Value::Char(l), Value::Char(r)) => Ok(Value::Boolean(l > r)),
            (l, r) => Err(format!("Cannot compare {} and {}", l, r)),
        }
    }

    fn lesser_equal(&self, right: &Value) -> Result<Value, String> {
        match (self, right) {
            (Value::Number(l), Value::Number(r)) => Ok(Value::Boolean(l <= r)),
            (Value::String(l), Value::String(r)) => Ok(Value::Boolean(l <= r)),
            (Value::Char(l), Value::Char(r)) => Ok(Value::Boolean(l <= r)),
            (l, r) => Err(format!("Cannot compare {} and {}", l, r)),
        }
    }

    fn lesser_than(&self, right: &Value) -> Result<Value, String> {
        match (self, right) {
            (Value::Number(l), Value::Number(r)) => Ok(Value::Boolean(l < r)),
            (Value::String(l), Value::String(r)) => Ok(Value::Boolean(l < r)),
            (Value::Char(l), Value::Char(r)) => Ok(Value::Boolean(l < r)),
            (l, r) => Err(format!("Cannot compare {} and {}", l, r)),
        }
    }
}

pub trait UnaryCompute {
    fn compute(&self, env: &mut dyn Environment, expr: &Expr) -> Result<Value, String>;
}

impl UnaryCompute for UnaryOp {
    fn compute(&self, env: &mut dyn Environment, expr: &Expr) -> Result<Value, String> {
        Ok(match self {
            UnaryOp::Minus => Value::Number(-expr.evaluate(env)?.int()?),
            UnaryOp::Not => Value::Boolean(!expr.evaluate(env)?.truth()?),
            UnaryOp::Len => Value::Number(expr.evaluate(env)?.len()?),
        })
    }
}

pub trait BinaryCompute {
    fn compute(
        &self,
        env: &mut dyn Environment,
        left: &Expr,
        right: &Expr,
    ) -> Result<Value, String>;
}

// add visitor to match types
impl BinaryCompute for BinaryOp {
    fn compute(
        &self,
        env: &mut dyn Environment,
        left: &Expr,
        right: &Expr,
    ) -> Result<Value, String> {
        match self {
            BinaryOp::Addition => left.evaluate(env)?.add(&right.evaluate(env)?),
            BinaryOp::Subtraction => left.evaluate(env)?.sub(&right.evaluate(env)?),
            BinaryOp::Multiplication => left.evaluate(env)?.mult(&right.evaluate(env)?),
            BinaryOp::Division => left.evaluate(env)?.div(&right.evaluate(env)?),
            BinaryOp::Modulo => left.evaluate(env)?.modulo(&right.evaluate(env)?),
            BinaryOp::Equal => left.evaluate(env)?.equal(&right.evaluate(env)?),
            BinaryOp::NotEqual => left.evaluate(env)?.not_equal(&right.evaluate(env)?),
            BinaryOp::GreaterEqual => left.evaluate(env)?.greater_equal(&right.evaluate(env)?),
            BinaryOp::GreaterThan => left.evaluate(env)?.greater_than(&right.evaluate(env)?),
            BinaryOp::LesserEqual => left.evaluate(env)?.lesser_equal(&right.evaluate(env)?),
            BinaryOp::LesserThan => left.evaluate(env)?.lesser_than(&right.evaluate(env)?),
            BinaryOp::And => Ok(Value::Boolean(
                left.evaluate(env)?.truth()? && right.evaluate(env)?.truth()?,
            )),
            BinaryOp::Or => Ok(Value::Boolean(
                left.evaluate(env)?.truth()? || right.evaluate(env)?.truth()?,
            )),
        }
    }
}
