use std::rc::Rc;

use crate::parser::syntax::Value;

use std::collections::HashMap;

pub trait Environment {
    fn create(&mut self, name: &str, node: Value) -> bool;
    fn assign(&mut self, name: &str, node: Value) -> Option<Value>;
    fn get(&self, name: &str) -> Option<Value>;
    fn recursive_to_string(&self, level: usize) -> (String, usize);
    fn to_string(&self) -> String {
        let (result, _) = self.recursive_to_string(0);
        result
    }
}

#[derive(Default, Debug)]
pub struct GlobalScope {
    global: HashMap<String, Value>,
}

impl Environment for GlobalScope {
    fn create(&mut self, name: &str, node: Value) -> bool {
        match self.global.contains_key(name) {
            true => false,
            false => {
                self.global.insert(name.to_string(), node);
                true
            }
        }
    }

    fn assign(&mut self, name: &str, node: Value) -> Option<Value> {
        match self.global.contains_key(name) {
            true => self.global.insert(name.to_string(), node),
            false => unreachable!("variable {} not found on assign, should never happen because it is checked before computing the value", name), 
        }
    }

    fn get(&self, name: &str) -> Option<Value> {
        self.global.get(name).cloned()
    }

    fn recursive_to_string(&self, level: usize) -> (String, usize) {
        let mut result = String::new();
        result.push_str("global:\n");
        for (key, val) in self.global.iter() {
            result.push_str(&format!("  var: {} => val: {}\n", key, val));
        }
        (result, level)
    }
}

pub struct LocalScope<'a> {
    name: String,
    parent: &'a mut dyn Environment,
    vars: HashMap<String, Value>,
}

impl LocalScope<'_> {
    pub fn new_function<'a>(
        parent: &'a mut (dyn Environment + 'a),
        name: &str,
        arg_names: &Rc<Vec<String>>,
        args: &[Value],
    ) -> LocalScope<'a> {
        let mut vars = HashMap::new();
        for (arg_name, arg) in arg_names.iter().zip(args.iter()) {
            vars.insert(arg_name.clone(), arg.clone());
        }
        let name = format!("{}({})", name, arg_names.join(", "));
        LocalScope { name, parent, vars }
    }

    pub fn new_block<'a>(parent: &'a mut (dyn Environment + 'a)) -> LocalScope<'a> {
        LocalScope {
            name: "block".to_string(),
            parent,
            vars: HashMap::new(),
        }
    }
}

impl<'a> Environment for LocalScope<'a> {
    fn create(&mut self, name: &str, node: Value) -> bool {
        match self.vars.contains_key(name) {
            true => false,
            false => {
                self.vars.insert(name.to_string(), node);
                true
            }
        }
    }

    fn assign(&mut self, name: &str, node: Value) -> Option<Value> {
        match self.vars.contains_key(name) {
            true => self.vars.insert(name.to_string(), node),
            false => self.parent.assign(name, node),
        }
    }

    fn get(&self, name: &str) -> Option<Value> {
        match self.vars.get(name) {
            Some(value) => Some(value.clone()),
            None => self.parent.get(name),
        }
    }

    fn recursive_to_string(&self, level: usize) -> (String, usize) {
        let (mut result, level_deep) = self.parent.recursive_to_string(level + 1);
        let indentation = "  ".repeat(level_deep - level);
        result.push_str(&format!("{}{}:\n", indentation, self.name));
        for (key, val) in self.vars.iter() {
            result.push_str(&format!("{}  var: {} => val: {}\n", indentation, key, val));
        }
        (result, level_deep)
    }
}
