use std::{io::Read, rc::Rc};

use super::{
    compute::{BinaryCompute, UnaryCompute},
    environment::{Environment, LocalScope},
    machine::{ControlFlow, Execution},
};
use crate::parser::syntax::{Expr, Value};

#[allow(clippy::len_without_is_empty)]
pub trait EvaluateValue {
    fn int(&self) -> Result<i64, String>;
    fn truth(&self) -> Result<bool, String>;
    fn str(&self) -> Result<Rc<Vec<u8>>, String>;
    fn char(&self) -> Result<u8, String>;
    fn len(&self) -> Result<i64, String>;
}

impl EvaluateValue for Value {
    fn int(&self) -> Result<i64, String> {
        match self {
            Value::Number(n) => Ok(*n),
            _ => Err(format!("number intended here, not {}", self.get_type())),
        }
    }
    fn truth(&self) -> Result<bool, String> {
        match self {
            Value::Boolean(b) => Ok(*b),
            Value::Number(value) => Ok(*value != 0),
            Value::String(string) => Ok(!string.is_empty()),
            Value::Char(c) => Ok(*c != b'\0'),
            _ => Err(format!("boolean intended here, not {}", self.get_type())),
        }
    }
    fn str(&self) -> Result<Rc<Vec<u8>>, String> {
        match self {
            Value::String(s) => Ok(s.clone()),
            _ => Err(format!("string intended here, not {}", self.get_type())),
        }
    }
    fn char(&self) -> Result<u8, String> {
        match self {
            Value::Char(c) => Ok(*c),
            _ => Err(format!("char intended here, not {}", self.get_type())),
        }
    }
    fn len(&self) -> Result<i64, String> {
        match self {
            Value::String(string) => Ok(string.len() as i64),
            Value::Array(array) => Ok(array.len() as i64),
            _ => Err(format!(
                "string or array intended with len operator, not {}",
                self.get_type()
            )),
        }
    }
}

fn evaluate_variable(
    env: &mut dyn Environment,
    name: &str,
    i_expr: &Option<Box<Expr>>,
) -> Result<Value, String> {
    if let Some(i_expr) = i_expr {
        let i = i_expr.evaluate(env)?.int()?;
        match env.get(name) {
            Some(Value::String(string)) => {
                if i < 0 || i >= string.len() as i64 {
                    return Err(format!(
                        "Index {} out of bounds for string of length {}",
                        i,
                        string.len()
                    ));
                }
                Ok(Value::Char(string[i as usize]))
            }
            Some(Value::Array(array)) => {
                if i < 0 || i >= array.len() as i64 {
                    return Err(format!(
                        "Index {} out of bounds for array of length {}",
                        i,
                        array.len()
                    ));
                }
                Ok(array[i as usize].clone())
            }
            Some(_) => Err(format!("{} is not indexable", name)),
            _ => Err(format!("Variable {} not found", name)),
        }
    } else {
        env.get(name)
            .ok_or_else(|| format!("Variable {} not found", name))
    }
}

fn evaluate_call(env: &mut dyn Environment, name: &str, args: &[Expr]) -> Result<Value, String> {
    let function = env
        .get(name)
        .ok_or_else(|| format!("function {} not found", name))?;
    match function {
        Value::Function(ref arg_names, ref body) => {
            let args = args
                .iter()
                .map(|arg| arg.evaluate(env))
                .collect::<Result<Vec<Value>, String>>()?;
            if arg_names.len() != args.len() {
                return Err(format!(
                    "function {} takes {} arguments, found {}",
                    name,
                    arg_names.len(),
                    args.len()
                ));
            }
            let mut fn_env = LocalScope::new_function(env, name, arg_names, &args);
            // println!("{}", fn_env.to_string());
            if let ControlFlow::Return(value) = body.execute(&mut fn_env)? {
                Ok(value)
            } else {
                Ok(Value::None)
            }
        }
        _ => Err(format!("{} is not a function", name)),
    }
}

pub trait EvaluateExpr {
    fn evaluate(&self, env: &mut dyn Environment) -> Result<Value, String>;
}

impl EvaluateExpr for Expr {
    fn evaluate(&self, env: &mut dyn Environment) -> Result<Value, String> {
        match *self {
            Expr::Litteral(ref value) => Ok(value.clone()),
            Expr::Variable(ref name, ref i_expr) => evaluate_variable(env, name, i_expr),
            Expr::UnaryOp(ref term, ref op) => op.compute(env, term),
            Expr::BinaryOp(ref lhs, ref rhs, ref op) => op.compute(env, lhs, rhs),
            Expr::Call(ref name, ref args) => evaluate_call(env, name, args),
            Expr::Array(ref exprs) => {
                let values = exprs
                    .iter()
                    .map(|expr| expr.evaluate(env))
                    .collect::<Result<Vec<Value>, String>>()?;
                Ok(Value::Array(Rc::new(values)))
            }
            Expr::Read(ref name) => match name.evaluate(env)?.to_string().as_str() {
                "stdin" => {
                    let mut buffer = String::new();
                    std::io::stdin()
                        .read_line(&mut buffer)
                        .expect("Error in reading from stdin");
                    Ok(Value::String(Rc::new(buffer.into_bytes())))
                }
                path => {
                    let mut f = std::fs::File::open(path)
                        .map_err(|_| format!("Error in opening file {}", path))?;
                    let mut content = String::new();
                    f.read_to_string(&mut content)
                        .map_err(|_| format!("Error in reading file {}", path))?;
                    Ok(Value::String(Rc::new(content.into_bytes())))
                }
            },
        }
    }
}
