use std::{
    io::{stderr, stdout, Write},
    rc::Rc,
};

use super::{
    environment::{Environment, GlobalScope, LocalScope},
    evaluate::{EvaluateExpr, EvaluateValue},
};
use crate::parser::syntax::{Stat, Value, WriteMode};

// can be used to add continue/exit
pub enum ControlFlow {
    Next,
    Break,
    Continue,
    Return(Value),
}

pub trait Execution {
    fn execute(&self, env: &mut dyn Environment) -> Result<ControlFlow, String>;
}

impl Execution for Vec<Stat> {
    fn execute(&self, env: &mut dyn Environment) -> Result<ControlFlow, String> {
        let mut env = LocalScope::new_block(env);
        for stat in self {
            match stat.execute(&mut env)? {
                ControlFlow::Next => (),
                flow @ (ControlFlow::Break | ControlFlow::Continue) => return Ok(flow),
                ControlFlow::Return(value) => return Ok(ControlFlow::Return(value)),
            }
        }
        Ok(ControlFlow::Next)
    }
}

impl Execution for Stat {
    fn execute(&self, env: &mut dyn Environment) -> Result<ControlFlow, String> {
        match *self {
            Stat::Condition(ref cond, ref then, ref alt) => {
                if cond.evaluate(env)?.truth()? {
                    return then.execute(env);
                } else if let Some(stats) = alt {
                    return stats.execute(env);
                }
                Ok(ControlFlow::Next)
            }
            Stat::Loop(ref cond, ref body) => {
                while cond.evaluate(env)?.truth()? {
                    if let ControlFlow::Break = body.execute(env)? {
                        break;
                    }
                }
                Ok(ControlFlow::Next)
            }
            Stat::Assignment(ref name, ref i_expr, ref expr) => {
                if let Some(i_expr) = i_expr {
                    match env.get(name) {
                        Some(Value::String(string)) => {
                            let i = i_expr.evaluate(env)?.int()?;
                            if i < 0 || i >= string.len() as i64 {
                                return Err(format!(
                                    "Index {} out of bounds for string of length {}",
                                    i,
                                    string.len()
                                ));
                            }
                            let mut copy = (*string).clone();
                            copy[i as usize] = expr.evaluate(env)?.char()?;
                            env.assign(name, Value::String(Rc::new(copy)));
                        }
                        Some(Value::Array(array)) => {
                            let i = i_expr.evaluate(env)?.int()?;
                            if i < 0 || i >= array.len() as i64 {
                                return Err(format!(
                                    "Index {} out of bounds for array of length {}",
                                    i,
                                    array.len()
                                ));
                            }
                            let mut copy = (*array).clone();
                            copy[i as usize] = expr.evaluate(env)?;
                            env.assign(name, Value::Array(Rc::new(copy)));
                        }
                        Some(_) => return Err(format!("Variable {} not indexable", name)),
                        _ => return Err(format!("Variable {} not found", name)),
                    }
                } else {
                    if env.get(name).is_none() {
                        return Err(format!("Variable {} not found", name));
                    }
                    let value = expr.evaluate(env)?;
                    env.assign(name, value);
                }
                // println!("{}", env.to_string());
                Ok(ControlFlow::Next)
            }
            Stat::Break => Ok(ControlFlow::Break),
            Stat::Continue => Ok(ControlFlow::Continue),
            Stat::FunDefine(ref name, ref args, ref body) => {
                match env.create(name, Value::Function(args.clone(), body.clone())) {
                    true => {
                        // println!("{}", env.to_string());
                        Ok(ControlFlow::Next)
                    }
                    false => Err(format!("Function {} already defined", name)),
                }
            }
            Stat::VarDefine(ref name, ref expr) => {
                let value = if let Some(expr) = expr {
                    expr.evaluate(env)?
                } else {
                    Value::None
                };
                match env.create(name, value) {
                    true => {
                        // println!("{}", env.to_string());
                        Ok(ControlFlow::Next)
                    }
                    false => Err(format!("Variable {} already defined", name)),
                }
            }
            Stat::Return(ref expr) => Ok(ControlFlow::Return(expr.evaluate(env)?)),
            Stat::Print(ref expr) => {
                println!("{}", expr.evaluate(env)?);
                Ok(ControlFlow::Next)
            }
            Stat::Write(ref expr, ref mode, ref name) => {
                let value = expr.evaluate(env)?;
                let name = name.evaluate(env)?.to_string();
                match name.as_str() {
                    "stdout" => {
                        if let WriteMode::Overwrite = mode {
                            print!("\r{}", termion::clear::CurrentLine);
                        }
                        print!("{}", value);
                        stdout().flush().unwrap();
                        Ok(ControlFlow::Next)
                    }
                    "stderr" => {
                        if let WriteMode::Overwrite = mode {
                            eprint!("{}", termion::clear::CurrentLine);
                        }
                        eprint!("{}", value);
                        stderr().flush().unwrap();
                        Ok(ControlFlow::Next)
                    }
                    path => {
                        match mode {
                            WriteMode::Overwrite => {
                                let mut file = std::fs::File::create(path).map_err(|e| {
                                    format!("Error in creating file {}: {}", path, e)
                                })?;
                                write!(file, "{}", value).map_err(|e| {
                                    format!("Error in writing to file {}: {}", path, e)
                                })?;
                            }
                            WriteMode::Append => {
                                let mut file = std::fs::OpenOptions::new()
                                    .append(true)
                                    .open(path)
                                    .map_err(|e| {
                                        format!("Error in opening file {}: {}", path, e)
                                    })?;
                                write!(file, "{}", value).map_err(|e| {
                                    format!("Error in writing to file {}: {}", path, e)
                                })?;
                            }
                        }
                        Ok(ControlFlow::Next)
                    }
                }
            }
        }
    }
}

pub struct Machine {
    pub env: GlobalScope,
}

impl Default for Machine {
    fn default() -> Self {
        Self::new()
    }
}

impl Machine {
    pub fn new() -> Machine {
        Machine {
            env: GlobalScope::default(),
        }
    }

    pub fn new_with_env(env: GlobalScope) -> Machine {
        Machine { env }
    }

    pub fn run(&mut self, stats: Vec<Stat>) -> Result<i64, String> {
        for stat in stats {
            if let ControlFlow::Return(val) = stat.execute(&mut self.env)? {
                return val.int();
            }
        }
        Ok(0)
    }
}
