pub mod compute;
pub mod environment;
pub mod evaluate;
pub mod machine;

#[cfg(test)]
mod tests {
    use crate::parser::{ast::build_ast, syntax::Value};
    use std::{fs::File, io::Read, str};

    use super::{
        environment::{Environment, GlobalScope},
        machine::Machine,
    };

    fn run_file(path: &str) -> Result<GlobalScope, String> {
        let mut f = File::open(path).expect(&format!("file {} not found", path));
        let mut content = String::new();
        f.read_to_string(&mut content)
            .expect(&format!("Error in reading file {}", path));
        let ast = build_ast(&content).unwrap();
        let mut machine = Machine::new();
        match machine.run(ast) {
            Ok(_) => Ok(machine.env),
            Err(error) => Err(error),
        }
    }

    fn run_string(string: &str) -> Result<GlobalScope, String> {
        let ast = build_ast(string).unwrap();
        let mut machine = Machine::new();
        match machine.run(ast) {
            Ok(_) => Ok(machine.env),
            Err(error) => Err(error),
        }
    }

    #[test]
    fn test_define_assign() {
        let environment = run_file("test_programs/define_assign.duck").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert_eq!(x, true);
        let y = match environment.get("y") {
            Some(Value::Number(value)) => value,
            _ => panic!("y should be Value::Number"),
        };
        assert_eq!(y, 1);
    }

    #[test]
    fn test_unknown_variable() {
        let environment = run_file("test_programs/unknown_variable.duck");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Variable y not found")
    }

    #[test]
    fn test_double_define_var() {
        let environment = run_file("test_programs/double_define_var.duck");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Variable x already defined")
    }

    #[test]
    fn test_double_define_fn() {
        let environment = run_file("test_programs/double_define_fn.duck");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Function test already defined")
    }

    #[test]
    fn test_arithmetic_expr() {
        let environment = run_file("test_programs/arithmetic_expr.duck").unwrap();
        let x = match environment.get("x") {
            Some(Value::Number(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, 2);
        let y = match environment.get("y") {
            Some(Value::Number(value)) => value,
            _ => panic!("y should be Value::Number"),
        };
        assert_eq!(y, 3 + 4 * (x + 1));
        let z = match environment.get("z") {
            Some(Value::Number(value)) => value,
            _ => panic!("z should be Value::Number"),
        };
        assert_eq!(z, x % (-3 / 2));
    }

    #[test]
    fn test_arithmetic_add_nb_nb() {
        let environment = run_string("var x = 3 + 4;").unwrap();
        let x = match environment.get("x") {
            Some(Value::Number(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, 7);
    }

    #[test]
    fn test_arithmetic_add_nb_nb_overflow() {
        let environment = run_string("var x = 4611686018427387904 + 4611686018427387904;");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Integer overflow: 4611686018427387904 + 4611686018427387904"
        )
    }

    #[test]
    fn test_arithmetic_add_char_nb() {
        let environment = run_string("var x = 'a' + 3;").unwrap();
        let x = match environment.get("x") {
            Some(Value::Char(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, b'a' + 3);
    }

    #[test]
    fn test_arithmetic_add_char_nb_overflow() {
        let environment = run_string("var x = 'z' + 200;");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Char overflow: z + 200")
    }

    #[test]
    fn test_arithmetic_add_nb_char() {
        let environment = run_string("var x = 3 + 'a';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Number(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, 3 + b'a' as i64);
    }

    #[test]
    fn test_arithmetic_add_nb_char_overflow() {
        let environment = run_string("var x = 9223372036854775807 + 'z';");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Integer overflow: 9223372036854775807 + z"
        )
    }

    #[test]
    fn test_arithmetic_add_char_char() {
        let environment = run_string("var x = 'a' + 'b';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Char(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, b'a' + b'b');
    }

    #[test]
    fn test_arithmetic_add_str_str() {
        let environment = run_string("var x = \"hello\" + \"world\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::String(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(str::from_utf8(&x).unwrap(), "helloworld");
    }

    #[test]
    fn test_arithmetic_add_str_char() {
        let environment = run_string("var x = \"hello\" + 'a';").unwrap();
        let x = match environment.get("x") {
            Some(Value::String(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(str::from_utf8(&x).unwrap(), "helloa");
    }

    #[test]
    fn test_arithmetic_add_char_str() {
        let environment = run_string("var x = 'a' + \"hello\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::String(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(str::from_utf8(&x).unwrap(), "ahello");
    }

    #[test]
    fn test_arithmetic_add_nb_str() {
        let environment = run_string("var x = 3 + \"hello\";");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot add number and string")
    }

    #[test]
    fn test_arithmetic_add_bool_str() {
        let environment = run_string("var x = true + \"hello\";");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot add boolean and string")
    }

    #[test]
    fn test_arithmetic_add_nb_bool() {
        let environment = run_string("var x = 3 + true;");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot add number and boolean")
    }

    #[test]
    fn test_arithmetic_add_bool_bool() {
        let environment = run_string("var x = true + false;");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot add boolean and boolean")
    }

    #[test]
    fn test_arithmetic_sub_nb_nb() {
        let environment = run_string("var x = 3 - 4;").unwrap();
        let x = match environment.get("x") {
            Some(Value::Number(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, -1);
    }

    #[test]
    fn test_arithmetic_sub_nb_nb_overflow() {
        let environment = run_string("var x = -4611686018427387904 - 4611686018427387905;");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Integer overflow: -4611686018427387904 - 4611686018427387905"
        )
    }

    #[test]
    fn test_arithmetic_sub_char_nb() {
        let environment = run_string("var x = 'a' - 3;").unwrap();
        let x = match environment.get("x") {
            Some(Value::Char(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, b'a' - 3);
    }

    #[test]
    fn test_arithmetic_sub_char_nb_overflow() {
        let environment = run_string("var x = 'a' - 200;");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Char overflow: a - 200")
    }

    #[test]
    fn test_arithmetic_sub_nb_char() {
        let environment = run_string("var x = 3 - 'a';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Number(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, 3 - b'a' as i64);
    }

    #[test]
    fn test_arithmetic_sub_nb_char_overflow() {
        let environment = run_string("var x = -9223372036854775807 - 'z';");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Integer overflow: -9223372036854775807 - z"
        )
    }

    #[test]
    fn test_arithmetic_sub_char_char() {
        let environment = run_string("var x = 'b' - 'a';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Char(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, b'b' - b'a');
    }

    #[test]
    fn test_arithmetic_sub_char_char_overflow() {
        let environment = run_string("var x = 'a' - 'z';");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Char overflow: a - z")
    }

    #[test]
    fn test_arithmetic_sub_str_str() {
        let environment = run_string("var x = \"hello\" - \"world\";");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Cannot subtract string and string"
        )
    }

    #[test]
    fn test_arithmetic_sub_str_char() {
        let environment = run_string("var x = \"hello\" - 'a';");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot subtract string and char")
    }

    #[test]
    fn test_arithmetic_sub_nb_str() {
        let environment = run_string("var x = 3 - \"hello\";");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Cannot subtract number and string"
        )
    }

    #[test]
    fn test_arithmetic_sub_bool_str() {
        let environment = run_string("var x = true - \"hello\";");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Cannot subtract boolean and string"
        )
    }

    #[test]
    fn test_arithmetic_sub_nb_bool() {
        let environment = run_string("var x = 3 - true;");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Cannot subtract number and boolean"
        )
    }

    #[test]
    fn test_arithmetic_sub_bool_bool() {
        let environment = run_string("var x = true - false;");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Cannot subtract boolean and boolean"
        )
    }

    #[test]
    fn test_arithmetic_mul_nb_nb() {
        let environment = run_string("var x = 3 * 4;").unwrap();
        let x = match environment.get("x") {
            Some(Value::Number(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, 12);
    }

    #[test]
    fn test_arithmetic_mul_nb_nb_overflow() {
        let environment = run_string("var x = 4611686018427387904 * 2;");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Integer overflow: 4611686018427387904 * 2"
        )
    }

    #[test]
    fn test_arithmetic_mul_char_nb() {
        let environment = run_string("var x = 'a' * 3;").unwrap();
        let x = match environment.get("x") {
            Some(Value::String(value)) => value,
            _ => panic!("x should be Value::String"),
        };
        assert_eq!(str::from_utf8(&x).unwrap(), "aaa");
    }

    #[test]
    fn test_arithmetic_mul_nb_char() {
        let environment = run_string("var x = 3 * 'a';").unwrap();
        let x = match environment.get("x") {
            Some(Value::String(value)) => value,
            _ => panic!("x should be Value::String"),
        };
        assert_eq!(str::from_utf8(&x).unwrap(), "aaa");
    }

    #[test]
    fn test_arithmetic_mul_char_char() {
        let environment = run_string("var x = 'a' * 'b';");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot multiply char and char")
    }

    #[test]
    fn test_arithmetic_mul_str_str() {
        let environment = run_string("var x = \"hello\" * \"world\";");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Cannot multiply string and string"
        )
    }

    #[test]
    fn test_arithmetic_mul_str_char() {
        let environment = run_string("var x = \"hello\" * 'a';");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot multiply string and char")
    }

    #[test]
    fn test_arithmetic_mul_nb_str() {
        let environment = run_string("var x = 3 * \"hello\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::String(value)) => value,
            _ => panic!("x should be Value::String"),
        };
        assert_eq!(str::from_utf8(&x).unwrap(), "hellohellohello");
    }

    #[test]
    fn test_arithmetic_mul_bool_str() {
        let environment = run_string("var x = true * \"hello\";");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Cannot multiply boolean and string"
        )
    }

    #[test]
    fn test_arithmetic_mul_nb_bool() {
        let environment = run_string("var x = 3 * true;");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Cannot multiply number and boolean"
        )
    }

    #[test]
    fn test_arithmetic_mul_bool_bool() {
        let environment = run_string("var x = true * false;");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Cannot multiply boolean and boolean"
        )
    }

    #[test]
    fn test_arithmetic_div_nb_nb() {
        let environment = run_string("var x = 3 / 4;").unwrap();
        let x = match environment.get("x") {
            Some(Value::Number(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, 0);
    }

    #[test]
    fn test_arithmetic_div_char_nb() {
        let environment = run_string("var x = 'a' / 3;");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot divide char and number")
    }

    #[test]
    fn test_arithmetic_div_char_char() {
        let environment = run_string("var x = 'a' / 'b';");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot divide char and char")
    }

    #[test]
    fn test_arithmetic_div_str_str() {
        let environment = run_string("var x = \"hello\" / \"world\";");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot divide string and string")
    }

    #[test]
    fn test_arithmetic_div_str_char() {
        let environment = run_string("var x = \"hello\" / 'a';");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot divide string and char")
    }

    #[test]
    fn test_arithmetic_div_nb_str() {
        let environment = run_string("var x = 3 / \"hello\";");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot divide number and string")
    }

    #[test]
    fn test_arithmetic_div_bool_str() {
        let environment = run_string("var x = true / \"hello\";");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot divide boolean and string")
    }

    #[test]
    fn test_arithmetic_div_nb_bool() {
        let environment = run_string("var x = 3 / true;");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot divide number and boolean")
    }

    #[test]
    fn test_arithmetic_div_bool_bool() {
        let environment = run_string("var x = true / false;");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Cannot divide boolean and boolean"
        )
    }

    #[test]
    fn test_arithmetic_error_div_zero() {
        let environment = run_string("var x = 3 / 0;");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "attempt to divide by zero")
    }

    #[test]
    fn test_arithmetic_mod_nb_nb() {
        let environment = run_string("var x = 3 % 4;").unwrap();
        let x = match environment.get("x") {
            Some(Value::Number(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, 3);
    }

    #[test]
    fn test_arithmetic_mod_char_nb() {
        let environment = run_string("var x = 'a' % 3;");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot modulo char and number")
    }

    #[test]
    fn test_arithmetic_mod_char_char() {
        let environment = run_string("var x = 'a' % 'b';");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot modulo char and char")
    }

    #[test]
    fn test_arithmetic_mod_str_str() {
        let environment = run_string("var x = \"hello\" % \"world\";");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot modulo string and string")
    }

    #[test]
    fn test_arithmetic_mod_str_char() {
        let environment = run_string("var x = \"hello\" % 'a';");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot modulo string and char")
    }

    #[test]
    fn test_arithmetic_mod_nb_str() {
        let environment = run_string("var x = 3 % \"hello\";");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot modulo number and string")
    }

    #[test]
    fn test_arithmetic_mod_bool_str() {
        let environment = run_string("var x = true % \"hello\";");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot modulo boolean and string")
    }

    #[test]
    fn test_arithmetic_mod_nb_bool() {
        let environment = run_string("var x = 3 % true;");
        assert!(environment.is_err());
        assert_eq!(environment.unwrap_err(), "Cannot modulo number and boolean")
    }

    #[test]
    fn test_arithmetic_mod_bool_bool() {
        let environment = run_string("var x = true % false;");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "Cannot modulo boolean and boolean"
        )
    }

    #[test]
    fn test_arithmetic_error_mod_zero() {
        let environment = run_string("var x = 3 % 0;");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "attempt to calculate the remainder with a divisor of zero"
        )
    }

    #[test]
    fn test_boolean_expr() {
        let environment = run_file("test_programs/boolean_expr.duck").unwrap();
        let a = match environment.get("a") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("a should be Value::Boolean"),
        };
        assert_eq!(a, (false || false));
        let b = match environment.get("b") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("b should be Value::Boolean"),
        };
        assert_eq!(b, (true || false));
        let c = match environment.get("c") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("c should be Value::Boolean"),
        };
        assert_eq!(c, (false || true));
        let d = match environment.get("d") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("d should be Value::Boolean"),
        };
        assert_eq!(d, (true || true));
        let e = match environment.get("e") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("e should be Value::Boolean"),
        };
        assert_eq!(e, (false && false));
        let f = match environment.get("f") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("f should be Value::Boolean"),
        };
        assert_eq!(f, (true && false));
        let g = match environment.get("g") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("g should be Value::Boolean"),
        };
        assert_eq!(g, (false && true));
        let h = match environment.get("h") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("h should be Value::Boolean"),
        };
        assert_eq!(h, (true && true));
        let i = match environment.get("i") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("i should be Value::Boolean"),
        };
        assert_eq!(i, !true);
        let j = match environment.get("j") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("j should be Value::Boolean"),
        };
        assert_eq!(j, !false);
    }

    #[test]
    fn test_boolean_expr_nb() {
        let environment = run_string("var x = 3 || 4;").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = 0 || 4;").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = 3 && 4;").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = 0 && 4;").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(!x);
    }

    #[test]
    fn test_boolean_expr_char() {
        let environment = run_string("var x = 'a' || 'b';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = '\\0' || 'b';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = 'a' && 'b';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = '\\0' && 'b';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(!x);
    }

    #[test]
    fn test_boolean_expr_str() {
        let environment = run_string("var x = \"hello\" || \"world\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = \"\" || \"world\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = \"hello\" && \"world\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = \"\" && \"world\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(!x);
    }

    #[test]
    fn test_compare_expr() {
        let environment = run_file("test_programs/compare_expr.duck").unwrap();
        let a = match environment.get("a") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("a should be Value::Boolean"),
        };
        assert_eq!(a, (1 < 2));
        let b = match environment.get("b") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("b should be Value::Boolean"),
        };
        assert_eq!(b, (2 < 1));
        let c = match environment.get("c") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("c should be Value::Boolean"),
        };
        assert_eq!(c, (2 < 2));
        let d = match environment.get("d") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("d should be Value::Boolean"),
        };
        assert_eq!(d, (1 > 2));
        let e = match environment.get("e") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("e should be Value::Boolean"),
        };
        assert_eq!(e, (2 > 1));
        let f = match environment.get("f") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("f should be Value::Boolean"),
        };
        assert_eq!(f, (2 > 2));
        let g = match environment.get("g") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("g should be Value::Boolean"),
        };
        assert_eq!(g, (1 <= 2));
        let h = match environment.get("h") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("h should be Value::Boolean"),
        };
        assert_eq!(h, (2 <= 1));
        let i = match environment.get("i") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("i should be Value::Boolean"),
        };
        assert_eq!(i, (2 <= 2));
        let j = match environment.get("j") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("j should be Value::Boolean"),
        };
        assert_eq!(j, (1 >= 2));
        let k = match environment.get("k") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("k should be Value::Boolean"),
        };
        assert_eq!(k, (2 >= 1));
        let l = match environment.get("l") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("l should be Value::Boolean"),
        };
        assert_eq!(l, (2 >= 2));
        let m = match environment.get("m") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("m should be Value::Boolean"),
        };
        assert_eq!(m, (1 == 2));
        let n = match environment.get("n") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("n should be Value::Boolean"),
        };
        assert_eq!(n, (2 == 2));
        let o = match environment.get("o") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("o should be Value::Boolean"),
        };
        assert_eq!(o, (1 != 2));
        let p = match environment.get("p") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("p should be Value::Boolean"),
        };
        assert_eq!(p, (2 != 2));
    }

    #[test]
    fn test_compare_expr_char() {
        let environment = run_string("var x = 'a' < 'b';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = 'b' < 'a';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(!x);
        let environment = run_string("var x = 'a' < 'a';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(!x);
        let environment = run_string("var x = 'a' > 'b';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(!x);
        let environment = run_string("var x = 'b' > 'a';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = 'a' > 'a';").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(!x);
    }

    #[test]
    fn test_compare_expr_str() {
        let environment = run_string("var x = \"hello\" < \"world\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = \"world\" < \"hello\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(!x);
        let environment = run_string("var x = \"hello\" < \"hello\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(!x);
        let environment = run_string("var x = \"hello\" > \"world\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(!x);
        let environment = run_string("var x = \"world\" > \"hello\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(x);
        let environment = run_string("var x = \"hello\" > \"hello\";").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert!(!x);
    }

    #[test]
    fn test_conditional() {
        let environment = run_file("test_programs/conditional.duck").unwrap();
        let x = match environment.get("x") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("x should be Value::Boolean"),
        };
        assert_eq!(x, true);
        let y = match environment.get("y") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("y should be Value::Boolean"),
        };
        assert_eq!(y, false);
        let z = match environment.get("z") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("z should be Value::Boolean"),
        };
        assert_eq!(z, true);
        let a = match environment.get("a") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("a should be Value::Boolean"),
        };
        assert_eq!(a, true);
        let b = match environment.get("b") {
            Some(Value::Boolean(value)) => value,
            _ => panic!("b should be Value::Boolean"),
        };
        assert_eq!(b, true);
    }

    #[test]
    fn test_while_loop() {
        let environment = run_file("test_programs/while_loop.duck").unwrap();
        let x = match environment.get("x") {
            Some(Value::Number(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, 5);
        let y = match environment.get("y") {
            Some(Value::Number(value)) => value,
            _ => panic!("y should be Value::Number"),
        };
        assert_eq!(y, 5);
        let z = match environment.get("z") {
            Some(Value::Number(value)) => value,
            _ => panic!("z should be Value::Number"),
        };
        assert_eq!(z, 2);
    }

    #[test]
    fn test_functions() {
        let environment = run_file("test_programs/functions.duck").unwrap();
        let x = match environment.get("x") {
            Some(Value::Number(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, 6765);
        let y = match environment.get("y") {
            Some(Value::Number(value)) => value,
            _ => panic!("y should be Value::Number"),
        };
        assert_eq!(y, 2);
        let z = match environment.get("z") {
            Some(Value::Number(value)) => value,
            _ => panic!("z should be Value::Number"),
        };
        assert_eq!(z, 0);
    }

    #[test]
    fn test_string_char() {
        let environment = run_file("test_programs/string_char.duck").unwrap();
        let x = match environment.get("x") {
            Some(Value::String(value)) => value,
            _ => panic!("x should be Value::String"),
        };
        assert_eq!(str::from_utf8(&x).unwrap(), "hallo ");
        let y = match environment.get("y") {
            Some(Value::String(value)) => value,
            _ => panic!("y should be Value::String"),
        };
        assert_eq!(str::from_utf8(&y).unwrap(), "world");
        let z = match environment.get("z") {
            Some(Value::String(value)) => value,
            _ => panic!("z should be Value::String"),
        };
        assert_eq!(str::from_utf8(&z).unwrap(), "hello world!");
        let a = match environment.get("a") {
            Some(Value::String(value)) => value,
            _ => panic!("a should be Value::String"),
        };
        assert_eq!(str::from_utf8(&a).unwrap(), "!!!");
        let b = match environment.get("b") {
            Some(Value::Char(value)) => value,
            _ => panic!("b should be Value::Char"),
        };
        assert_eq!(b, b'l');
        let c = match environment.get("c") {
            Some(Value::String(value)) => value,
            _ => panic!("a should be Value::String"),
        };
        assert_eq!(str::from_utf8(&c).unwrap(), "world world world ");
        let d = match environment.get("d") {
            Some(Value::String(value)) => value,
            _ => panic!("d should be Value::String"),
        };
        assert_eq!(str::from_utf8(&d).unwrap(), "{\n\ttest: \"test\"\n}\n");
        let e = match environment.get("e") {
            Some(Value::Char(value)) => value,
            _ => panic!("b should be Value::Char"),
        };
        assert_eq!(e, b'\'');
        let f = match environment.get("f") {
            Some(Value::Number(value)) => value,
            _ => panic!("b should be Number::Char"),
        };
        assert_eq!(f, 12);
    }

    #[test]
    fn test_local_block_env() {
        let environment = run_file("test_programs/local_block_env.duck").unwrap();
        let x = match environment.get("x") {
            Some(Value::Number(value)) => value,
            _ => panic!("x should be Value::Number"),
        };
        assert_eq!(x, 10);
        let y = match environment.get("y") {
            Some(Value::Number(value)) => value,
            _ => panic!("y should be Value::Number"),
        };
        assert_eq!(y, 0);
        let z = match environment.get("z") {
            Some(Value::Number(value)) => value,
            _ => panic!("z should be Value::Number"),
        };
        assert_eq!(z, 1);
        assert!(match environment.get("n") {
            Some(_) => false,
            None => true,
        });
    }

    #[test]
    fn test_arrays() {
        let environment = run_file("test_programs/arrays.duck").unwrap();
        let a = match environment.get("a") {
            Some(value) => match value {
                Value::Array(_) => value,
                _ => panic!("a should be Value::Array"),
            },
            _ => panic!("a should be defined"),
        };
        assert_eq!(format!("{}", a), "[1, 2, 3, 4, 5, 6, 7, 8, 9]");
        let b = match environment.get("b") {
            Some(value) => match value {
                Value::Array(_) => value,
                _ => panic!("y should be Value::Array"),
            },
            _ => panic!("y should be defined"),
        };
        assert_eq!(format!("{}", b), "[3, 10]");
        let c = match environment.get("c") {
            Some(value) => match value {
                Value::Array(_) => value,
                _ => panic!("c should be Value::Array"),
            },
            _ => panic!("c should be defined"),
        };
        assert_eq!(format!("{}", c), "[1, 2, 3, 4, 5, 6, 7, 8, 9, 3, 10]");
        let d = match environment.get("d") {
            Some(value) => match value {
                Value::Array(_) => value,
                _ => panic!("d should be Value::Array"),
            },
            _ => panic!("d should be defined"),
        };
        assert_eq!(format!("{}", d), "[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]");
        let e = match environment.get("e") {
            Some(value) => match value {
                Value::Array(_) => value,
                _ => panic!("e should be Value::Array"),
            },
            _ => panic!("e should be defined"),
        };
        assert_eq!(format!("{}", e), "[1, 4, 3]");
        let f = match environment.get("f") {
            Some(value) => match value {
                Value::Array(_) => value,
                _ => panic!("f should be Value::Array"),
            },
            _ => panic!("f should be defined"),
        };
        assert_eq!(format!("{}", f), "[2, 4, 6]");
    }

    #[test]
    fn test_len_not_array_str() {
        let environment = run_string("var x = len 2;");
        assert!(environment.is_err());
        assert_eq!(
            environment.unwrap_err(),
            "string or array intended with len operator, not number"
        )
    }
}
