use std::rc::Rc;

use pest::iterators::Pair;
use pest::pratt_parser::PrattParser;
use pest::Parser;

use super::{
    error::{parsing_error, renamed_parsing_error_rules},
    escape::{interpret_escaped_char, interpret_escaped_string, EscapeError},
    syntax::{BinaryOp, Expr, Stat, UnaryOp, Value, WriteMode},
};

#[derive(Parser)]
#[grammar = "parser/duck.pest"]
pub struct DuckParser;

lazy_static::lazy_static! {
    static ref PRATT_PARSER: PrattParser<Rule> = {
        use pest::pratt_parser::{Assoc::*, Op};
        use Rule::*;

        // Precedence is defined lowest to highest
        PrattParser::new()
            // Addition and subtract have equal precedence
            .op(Op::infix(op_or, Left) | Op::infix(op_and, Left))
            .op(Op::infix(op_lt, Left) | Op::infix(op_gt, Left) | Op::infix(op_eq, Left) | Op::infix(op_ge, Left) | Op::infix(op_le, Left) | Op::infix(op_ne, Left))
            .op(Op::infix(op_add, Left) | Op::infix(op_sub, Left))
            .op(Op::infix(op_mul, Left) | Op::infix(op_div, Left) | Op::infix(op_mod, Left))
            .op(Op::prefix(op_neg) | Op::prefix(op_not) | Op::prefix(op_len))
    };
}

fn build_call(pair: Pair<Rule>) -> Result<Expr, String> {
    let mut inner = pair.into_inner();
    let name = inner.next().unwrap().as_str().to_string();
    let args = inner.next().unwrap().into_inner();
    let arg_exprs = args
        .into_iter()
        .map(|arg| build_expr(arg).unwrap())
        .collect();
    Ok(Expr::Call(name, arg_exprs))
}

fn build_variable(pair: Pair<Rule>) -> Result<(String, Option<Box<Expr>>), String> {
    let mut inner = pair.into_inner();
    let name = inner.next().unwrap().as_str().to_string();
    let index = match inner.next() {
        Some(index) => Some(Box::new(build_expr(index)?)),
        None => None,
    };
    Ok((name, index))
}

fn build_expr(pair: Pair<Rule>) -> Result<Expr, String> {
    PRATT_PARSER
        .map_primary(|primary| match primary.as_rule() {
            Rule::term | Rule::expr => build_expr(primary),
            Rule::number => Ok(Expr::Litteral(Value::number(primary.as_str())?)),
            Rule::boolean => Ok(Expr::Litteral(Value::boolean(primary.as_str())?)),
            Rule::variable => {
                let (name, index) = build_variable(primary)?;
                Ok(Expr::Variable(name, index))
            }
            Rule::call => build_call(primary),
            Rule::string => {
                let litteral = primary.as_str();
                let value = match interpret_escaped_string(&litteral[1..litteral.len() - 1]) {
                    Ok(value) => value,
                    Err(EscapeError::EscapeAtEndOfString) => {
                        return Err(parsing_error(primary, "escape character at end of string"))
                    }
                    Err(EscapeError::InvalidEscapedChar(c)) => {
                        return Err(parsing_error(
                            primary,
                            &format!("invalid escape character {}", c),
                        ))
                    }
                };
                Ok(Expr::Litteral(Value::String(Rc::new(value))))
            }
            Rule::char => {
                let litteral = primary.as_str();
                let value = match interpret_escaped_char(&litteral[1..litteral.len() - 1]) {
                    Ok(value) => value,
                    Err(EscapeError::EscapeAtEndOfString) => {
                        return Err(parsing_error(primary, "escape character at end of string"))
                    }
                    Err(EscapeError::InvalidEscapedChar(c)) => {
                        return Err(parsing_error(
                            primary,
                            &format!("invalid escape character {}", c),
                        ))
                    }
                };
                Ok(Expr::Litteral(Value::Char(value)))
            }
            Rule::array => {
                let mut exprs = Vec::new();
                for pair in primary.into_inner() {
                    exprs.push(build_expr(pair)?);
                }
                Ok(Expr::Array(exprs))
            }
            Rule::read => {
                let mut inner = primary.into_inner();
                let name = build_expr(inner.nth(1).unwrap())?;
                Ok(Expr::Read(Box::new(name)))
            }
            rule => Err(format!(
                "atom expected number, boolean or variable, found {:?}",
                rule
            )),
        })
        .map_infix(|lhs, op, rhs| match (lhs, op.as_rule(), rhs) {
            (Err(err), _, _) | (_, _, Err(err)) => Err(err),
            (Ok(lhs), Rule::op_or, Ok(rhs)) => {
                Ok(Expr::BinaryOp(Box::new(lhs), Box::new(rhs), BinaryOp::Or))
            }
            (Ok(lhs), Rule::op_and, Ok(rhs)) => {
                Ok(Expr::BinaryOp(Box::new(lhs), Box::new(rhs), BinaryOp::And))
            }
            (Ok(lhs), Rule::op_lt, Ok(rhs)) => Ok(Expr::BinaryOp(
                Box::new(lhs),
                Box::new(rhs),
                BinaryOp::LesserThan,
            )),
            (Ok(lhs), Rule::op_gt, Ok(rhs)) => Ok(Expr::BinaryOp(
                Box::new(lhs),
                Box::new(rhs),
                BinaryOp::GreaterThan,
            )),
            (Ok(lhs), Rule::op_eq, Ok(rhs)) => Ok(Expr::BinaryOp(
                Box::new(lhs),
                Box::new(rhs),
                BinaryOp::Equal,
            )),
            (Ok(lhs), Rule::op_ge, Ok(rhs)) => Ok(Expr::BinaryOp(
                Box::new(lhs),
                Box::new(rhs),
                BinaryOp::GreaterEqual,
            )),
            (Ok(lhs), Rule::op_le, Ok(rhs)) => Ok(Expr::BinaryOp(
                Box::new(lhs),
                Box::new(rhs),
                BinaryOp::LesserEqual,
            )),
            (Ok(lhs), Rule::op_ne, Ok(rhs)) => Ok(Expr::BinaryOp(
                Box::new(lhs),
                Box::new(rhs),
                BinaryOp::NotEqual,
            )),
            (Ok(lhs), Rule::op_add, Ok(rhs)) => Ok(Expr::BinaryOp(
                Box::new(lhs),
                Box::new(rhs),
                BinaryOp::Addition,
            )),
            (Ok(lhs), Rule::op_sub, Ok(rhs)) => Ok(Expr::BinaryOp(
                Box::new(lhs),
                Box::new(rhs),
                BinaryOp::Subtraction,
            )),
            (Ok(lhs), Rule::op_mul, Ok(rhs)) => Ok(Expr::BinaryOp(
                Box::new(lhs),
                Box::new(rhs),
                BinaryOp::Multiplication,
            )),
            (Ok(lhs), Rule::op_div, Ok(rhs)) => Ok(Expr::BinaryOp(
                Box::new(lhs),
                Box::new(rhs),
                BinaryOp::Division,
            )),
            (Ok(lhs), Rule::op_mod, Ok(rhs)) => Ok(Expr::BinaryOp(
                Box::new(lhs),
                Box::new(rhs),
                BinaryOp::Modulo,
            )),
            (_, rule, _) => Err(format!(
                "binary operation expected operator, found {:?}",
                rule
            )),
        })
        .map_prefix(|op, rhs| match (op.as_rule(), rhs) {
            (_, Err(err)) => Err(err),
            (Rule::op_neg, Ok(rhs)) => Ok(Expr::UnaryOp(Box::new(rhs), UnaryOp::Minus)),
            (Rule::op_not, Ok(rhs)) => Ok(Expr::UnaryOp(Box::new(rhs), UnaryOp::Not)),
            (Rule::op_len, Ok(rhs)) => Ok(Expr::UnaryOp(Box::new(rhs), UnaryOp::Len)),
            (rule, _) => Err(format!(
                "unary operation expected operator, found {:?}",
                rule
            )),
        })
        .parse(pair.into_inner())
}

fn build_assign(pair: Pair<Rule>) -> Result<Stat, String> {
    let mut inner = pair.into_inner();
    let (name, index) = build_variable(inner.next().unwrap())?;
    let expr = build_expr(inner.nth(1).unwrap())?;
    Ok(Stat::Assignment(name, index, expr))
}

fn build_while_loop(pair: Pair<Rule>) -> Result<Stat, String> {
    let mut inner = pair.into_inner();
    let cond = build_expr(inner.nth(1).unwrap())?;
    let stat = build_stats(inner.next().unwrap(), true)?;
    Ok(Stat::Loop(cond, stat))
}

fn build_if_cond(pair: Pair<Rule>, in_loop: bool) -> Result<Stat, String> {
    let mut inner = pair.into_inner();
    let cond = build_expr(inner.nth(1).unwrap())?;
    let then = build_stats(inner.next().unwrap(), in_loop)?;
    match inner.next() {
        Some(pair) => match pair.as_rule() {
            Rule::else_keyword => Ok(Stat::Condition(
                cond,
                then,
                Some(build_stats(inner.next().unwrap(), in_loop)?),
            )),
            Rule::stat_elif => Ok(Stat::Condition(
                cond,
                then,
                Some(vec![build_if_cond(pair, in_loop)?]),
            )),
            rule => Err(format!("conditional statement expected, found {:?}", rule)),
        },
        None => Ok(Stat::Condition(cond, then, None)),
    }
}

fn build_define(pair: Pair<Rule>) -> Result<Stat, String> {
    let inner = pair.into_inner().next().unwrap();
    match inner.as_rule() {
        Rule::function_define => {
            let mut inner = inner.into_inner();
            let name = inner.nth(1).unwrap().as_str().to_string();
            let args = inner.next().unwrap().into_inner();
            let arg_names = args
                .into_iter()
                .map(|arg| arg.as_str().to_string())
                .collect();
            let stat = build_stats(inner.next().unwrap(), false)?;
            Ok(Stat::FunDefine(name, Rc::new(arg_names), Rc::new(stat)))
        }
        Rule::variable_define => {
            let mut inner = inner.into_inner();
            let name = inner.nth(1).unwrap().as_str().to_string();
            if inner.len() == 1 {
                return Ok(Stat::VarDefine(name, None));
            }
            let expr = build_expr(inner.nth(1).unwrap())?;
            Ok(Stat::VarDefine(name, Some(expr)))
        }
        rule => Err(format!("definition expected, found {:?}", rule)),
    }
}

fn build_return(pair: Pair<Rule>) -> Result<Stat, String> {
    let expr = build_expr(pair.into_inner().nth(1).unwrap())?;
    Ok(Stat::Return(expr))
}

fn build_print(pair: Pair<Rule>) -> Result<Stat, String> {
    let expr = build_expr(pair.into_inner().nth(1).unwrap())?;
    Ok(Stat::Print(expr))
}

fn build_write(pair: Pair<Rule>) -> Result<Stat, String> {
    let mut inner = pair.into_inner();
    let value = build_expr(inner.nth(1).unwrap())?;
    let mode = match inner.next().unwrap().as_rule() {
        Rule::to_keyword => WriteMode::Append,
        Rule::new_keyword => WriteMode::Overwrite,
        _ => unreachable!(),
    };
    let name = build_expr(inner.next().unwrap())?;
    Ok(Stat::Write(value, mode, name))
}

fn build_stat(pair: Pair<Rule>, in_loop: bool) -> Result<Stat, String> {
    match pair.as_rule() {
        Rule::stat_assign => build_assign(pair),
        Rule::stat_if | Rule::stat_elif => build_if_cond(pair, in_loop),
        Rule::stat_while => build_while_loop(pair),
        Rule::stat_break => {
            if in_loop {
                return Ok(Stat::Break);
            }
            Err(parsing_error(pair, "break keyword while not in loop"))
        }
        Rule::stat_continue => {
            if in_loop {
                return Ok(Stat::Continue);
            }
            Err(parsing_error(pair, "continue keyword while not in loop"))
        }
        Rule::stat_define => build_define(pair),
        Rule::stat_return => build_return(pair),
        Rule::stat_print => build_print(pair),
        Rule::stat_write => build_write(pair),
        rule => Err(format!("Statement expected, found {:?}", rule)),
    }
}

fn build_stats(pair: Pair<Rule>, in_loop: bool) -> Result<Vec<Stat>, String> {
    let inner = pair.into_inner();
    let mut stats = Vec::new();
    for pair in inner {
        stats.push(build_stat(pair, in_loop)?);
    }
    Ok(stats)
}

pub fn build_ast(content: &str) -> Result<Vec<Stat>, String> {
    let pair = match DuckParser::parse(Rule::program, content) {
        Ok(mut pairs) => pairs.next().unwrap(),
        Err(err) => {
            return Err(format!(
                "{}",
                err.renamed_rules(renamed_parsing_error_rules)
            ))
        }
    };
    build_stats(pair, false)
}
