#![allow(clippy::should_implement_trait)]
use std::{
    fmt::{self, Display, Formatter},
    rc::Rc,
    str,
};

#[derive(Clone, Copy, Debug)]
pub enum BinaryOp {
    Addition,
    Subtraction,
    Multiplication,
    Division,
    Modulo,
    LesserThan,
    GreaterThan,
    Equal,
    GreaterEqual,
    LesserEqual,
    NotEqual,
    Or,
    And,
}

impl Display for BinaryOp {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        match self {
            BinaryOp::Addition => write!(f, "+"),
            BinaryOp::Subtraction => write!(f, "-"),
            BinaryOp::Multiplication => write!(f, "*"),
            BinaryOp::Division => write!(f, "/"),
            BinaryOp::Modulo => write!(f, "%"),
            BinaryOp::LesserThan => write!(f, "<"),
            BinaryOp::GreaterThan => write!(f, ">"),
            BinaryOp::Equal => write!(f, "=="),
            BinaryOp::GreaterEqual => write!(f, ">="),
            BinaryOp::LesserEqual => write!(f, "<="),
            BinaryOp::NotEqual => write!(f, "!="),
            BinaryOp::Or => write!(f, "||"),
            BinaryOp::And => write!(f, "&&"),
        }
    }
}

#[derive(Clone, Debug)]
pub enum UnaryOp {
    Minus,
    Not,
    Len,
}

impl Display for UnaryOp {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        match self {
            UnaryOp::Minus => write!(f, "-"),
            UnaryOp::Not => write!(f, "!"),
            UnaryOp::Len => write!(f, "len "),
        }
    }
}

// Rc all the heavy stuff
#[derive(Clone, Debug)]
pub enum Value {
    Number(i64),
    Boolean(bool),
    Function(Rc<Vec<String>>, Rc<Vec<Stat>>),
    String(Rc<Vec<u8>>),
    Char(u8),
    Array(Rc<Vec<Value>>),
    None,
}

impl Value {
    pub fn number(raw: &str) -> Result<Value, String> {
        match raw.parse::<i64>() {
            Ok(value) => Ok(Value::Number(value)),
            Err(_) => Err(format!("{} is not a number", raw)),
        }
    }
    pub fn boolean(raw: &str) -> Result<Value, String> {
        match raw.parse::<bool>() {
            Ok(value) => Ok(Value::Boolean(value)),
            Err(_) => Err(format!("{} is not a boolean", raw)),
        }
    }
    pub fn get_type(&self) -> &'static str {
        match self {
            Value::Number(_) => "number",
            Value::Boolean(_) => "boolean",
            Value::Function(_, _) => "function",
            Value::String(_) => "string",
            Value::Char(_) => "char",
            Value::Array(_) => "array",
            Value::None => "none",
        }
    }
}

impl Display for Value {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        match self {
            Value::Number(n) => write!(f, "{}", n),
            Value::Boolean(b) => write!(f, "{}", b),
            Value::Function(args, _) => write!(f, "fn({})", args.join(", ")),
            Value::String(s) => unsafe { write!(f, "{}", str::from_utf8_unchecked(s)) },
            Value::Char(c) => write!(f, "{}", char::from(*c)),
            Value::Array(arr) => {
                write!(f, "[")?;
                let values = arr
                    .iter()
                    .map(|value| value.to_string())
                    .collect::<Vec<String>>();
                write!(f, "{}", values.join(", "))?;
                write!(f, "]")
            }
            Value::None => write!(f, "None"),
        }
    }
}

// box because of recursive definition
#[derive(Clone, Debug)]
pub enum Expr {
    Litteral(Value),
    Variable(String, Option<Box<Expr>>),
    BinaryOp(Box<Expr>, Box<Expr>, BinaryOp),
    UnaryOp(Box<Expr>, UnaryOp),
    Call(String, Vec<Expr>),
    Array(Vec<Expr>),
    Read(Box<Expr>),
}

impl Display for Expr {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        match self {
            Expr::Litteral(value) => write!(f, "{}", value),
            Expr::Variable(name, index) => {
                if let Some(index) = index {
                    write!(f, "{}[{}]", name, index)
                } else {
                    write!(f, "{}", name)
                }
            }
            Expr::BinaryOp(left, right, op) => write!(f, "{} {} {}", left, op, right),
            Expr::UnaryOp(term, op) => write!(f, "{}{}", op, term),
            Expr::Call(name, args) => {
                write!(f, "{}(", name)?;
                let exprs = args
                    .iter()
                    .map(|arg| arg.to_string())
                    .collect::<Vec<String>>();
                write!(f, "{}", exprs.join(", "))?;
                write!(f, ")")
            }
            Expr::Array(exprs) => {
                write!(f, "[")?;
                let exprs = exprs
                    .iter()
                    .map(|arg| arg.to_string())
                    .collect::<Vec<String>>();
                write!(f, "{}", exprs.join(", "))?;
                write!(f, "]")
            }
            Expr::Read(name) => write!(f, "read {}", name),
        }
    }
}

#[derive(Clone, Copy, Debug)]
pub enum WriteMode {
    Overwrite,
    Append,
}

impl Display for WriteMode {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        match self {
            WriteMode::Overwrite => write!(f, "new"),
            WriteMode::Append => write!(f, "to"),
        }
    }
}

#[derive(Clone, Debug)]
pub enum Stat {
    Condition(Expr, Vec<Stat>, Option<Vec<Stat>>),
    Loop(Expr, Vec<Stat>),
    Assignment(String, Option<Box<Expr>>, Expr),
    Break,
    Continue,
    FunDefine(String, Rc<Vec<String>>, Rc<Vec<Stat>>),
    VarDefine(String, Option<Expr>),
    Return(Expr),
    Print(Expr),
    Write(Expr, WriteMode, Expr),
}

impl Display for Stat {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        match self {
            Stat::Condition(cond, then, alt) => {
                write!(f, "if ({}) {{", cond)?;
                for stat in then {
                    write!(f, " {} ", stat)?;
                }
                write!(f, "}}")?;
                if let Some(alt_stats) = alt {
                    write!(f, " else {{")?;
                    for stat in alt_stats {
                        write!(f, " {} ", stat)?;
                    }
                    write!(f, "}}")?;
                }
                Ok(())
            }
            Stat::Loop(cond, body) => {
                write!(f, "while ({}) {{", cond)?;
                for stat in body {
                    write!(f, " {} ", stat)?;
                }
                write!(f, "}}")
            }
            Stat::Assignment(name, index, expr) => {
                if let Some(index) = index {
                    write!(f, "{}[{}] = {};", name, index, expr)
                } else {
                    write!(f, "{} = {};", name, expr)
                }
            }
            Stat::Break => write!(f, "break;"),
            Stat::Continue => write!(f, "continue;"),
            Stat::FunDefine(name, args, body) => {
                write!(f, "fn {}( ", name)?;
                write!(f, "{}", args.join(", "))?;
                write!(f, " ) {{")?;
                for statement in body.as_ref() {
                    write!(f, " {} ", statement)?;
                }
                write!(f, "}}")
            }
            Stat::VarDefine(name, expr) => match expr {
                Some(expr) => write!(f, "var {} = {};", name, expr),
                None => write!(f, "var {};", name),
            },
            Stat::Return(expr) => write!(f, "return {};", expr),
            Stat::Print(expr) => write!(f, "print {};", expr),
            Stat::Write(name, mode, expr) => write!(f, "write {} {} {}", name, mode, expr),
        }
    }
}
