use ducklang::{interpreter::machine::Machine, parser::ast::build_ast};

const EXAMPLE_CODE: &str = r#"/*
An example of program with most of
the supported features
*/
x = true;
z = 3 + 4 * (2 + 1); // good order of op
if (x) {
	y = 5;
	while (x) {
		if (!y && z % 2) {
			break;
		} else if (y == 3) { // never reached because of continue
			z = 5;
		} else if (y == 4) {
			y = y - 2;
			continue;
		}
		y = y - 1;
	}
} else {
	// never reached
	y = -8;
	x = true;
}"#;

pub struct MyEguiApp {
    code: String,
    machine: Machine,
    error: String,
}

impl MyEguiApp {
    pub fn new(_cc: &eframe::CreationContext<'_>) -> Self {
        Self {
            code: String::from(EXAMPLE_CODE),
            machine: Machine::new(),
            error: String::new(),
        }
    }
}

impl eframe::App for MyEguiApp {
    fn update(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        let _toolbar = self.render_toolbar(ctx);
        let _side_panel = self.render_side_panel(ctx);
        let _error_panel = self.render_error_panel(ctx);
        let _central_panel = self.render_central_panel(ctx);
    }
}

impl MyEguiApp {
    fn render_error_panel(&mut self, ctx: &egui::Context) -> egui::Rect {
        let error_panel = egui::TopBottomPanel::bottom("error_panel").show(ctx, |ui| {
            ui.heading("Error");
            ui.separator();
            ui.add(
                egui::TextEdit::multiline(&mut self.error.as_str()).desired_width(f32::INFINITY),
            );
        });
        error_panel.response.rect
    }

    fn render_central_panel(&mut self, ctx: &egui::Context) -> egui::Rect {
        let central_panel = egui::CentralPanel::default().show(ctx, |ui| {
            // calculate desired rows based on the height of the window
            let font_selection = egui::FontSelection::default();
            let font_id = font_selection.resolve(ui.style());
            let row_height = ui.fonts(|fonts| fonts.row_height(&font_id));
            let rows = (ui.available_size().y / row_height) as usize;
            ui.add(
                egui::TextEdit::multiline(&mut self.code)
                    .code_editor()
                    .desired_width(f32::INFINITY)
                    .desired_rows(rows),
            );
        });
        central_panel.response.rect
    }

    fn render_side_panel(&mut self, ctx: &egui::Context) -> egui::Rect {
        let side_panel = egui::SidePanel::left("side_panel").show(ctx, |ui| {
            ui.heading("Environment");
            ui.separator();
            // TODO fix for new features
            // for (name, value) in self.machine.env.global.iter() {
            //     ui.label(format!("{}: {}", name, value));
            // }
        });
        side_panel.response.rect
    }

    fn render_toolbar(&mut self, ctx: &egui::Context) -> egui::Rect {
        let mut frame = egui::Frame::side_top_panel(ctx.style().as_ref());
        frame.inner_margin.bottom += 1.5;
        let toolbar = egui::TopBottomPanel::top("toolbar")
            .frame(frame)
            .show(ctx, |ui| {
                ui.horizontal(|ui| {
                    ui.style_mut().spacing.interact_size.y *= 1.25;
                    ui.style_mut()
                        .text_styles
                        .get_mut(&egui::TextStyle::Button)
                        .unwrap()
                        .size *= 1.25;
                    if ui.button("Reset").clicked() {
                        self.reset();
                    }
                    if ui.button("Run").clicked() {
                        self.run();
                    }
                });
            });
        toolbar.response.rect
    }

    fn reset(&mut self) {
        self.machine = Machine::new();
    }

    fn run(&mut self) {
        match build_ast(&self.code) {
            Ok(ast) => match self.machine.run(ast) {
                Ok(_) => {
                    self.error = String::new();
                }
                Err(e) => {
                    self.error = e;
                }
            },
            Err(e) => {
                self.error = e;
            }
        }
    }
}

#[cfg(target_arch = "wasm32")]
fn main() {
    // Make sure panics are logged using `console.error`.
    console_error_panic_hook::set_once();
    tracing_wasm::set_as_global_default();

    let web_options = eframe::WebOptions::default();

    wasm_bindgen_futures::spawn_local(async {
        eframe::start_web(
            // this is the id of the `<canvas>` element we have
            // in our `index.html`
            "canvas",
            web_options,
            Box::new(|cc| Box::new(MyEguiApp::new(cc))),
        )
        .await
        .expect("failed to start eframe");
    });
}

#[cfg(not(target_arch = "wasm32"))]
fn main() {
    let options = eframe::NativeOptions {
        viewport: egui::ViewportBuilder::default()
            .with_inner_size([1280.0, 720.0])
            .with_resizable(true),
        ..Default::default()
    };
    let _run_native = eframe::run_native(
        "Duckscript Playground",
        options,
        Box::new(|cc| Box::new(MyEguiApp::new(cc))),
    );
}
